
void F77_NAME(propclusttrial) (float *, int *, double *, double *, double *, double *, int *,int *,int *, int *);
void F77_NAME(propclustaccel) (float *, int *, double *, double *, double *, double *, int *,int *,int *, int *);
void F77_NAME(propensityclustering) (float *, int *, double *, double *, double *, double *, 
                                     int *,int *,int *, int *);

void F77_NAME(propensitydecomposition) (float *, int *, double *, double *, double *, double *, int *,int *,int *);
void F77_NAME(propdecompaccel) (float *, int *, double *, double *, double *, double *, int *,int *,int *);

void F77_NAME(singleclusterupdate) (float *, double *, double *, double *, int *,int *);
void F77_NAME(speedtest1) (double *, int *);
void F77_NAME(speedtest2) (double *, int *);

